#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <string.h>
#include <time.h>
#include <sys/time.h>
#include <errno.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <stdarg.h>
#include <signal.h>
#include "version.h"


#define kStart 0
#define kOpenPort 1
#define kInitModem 2
#define kWaitRing 3
#define kAnswer 4
#define kPPPd 5
#define kHangup 6
#define kError 7
#define kFatalError 8

#define rError 0
#define rOK 1
#define rTimeout 2

#define newLine "\r"

#define initString "AT S7=45 S0=0 L1 V1 X4 &c1 E1 Q0"
#define cmdOK "OK"
#define cmdRING "RING"
#define cmdANSWER "ATA"
#define cmdCONNECT "CONNECT"
#define cmdHANGUP "ATH"

#define LOGFILE "/var/log/ras.log"

#define PPPD_START "#pppd="
#define INIT_START "#init="
#define DEV_START "/dev/"

#define PPPD_DEFAULT "/usr/sbin/pppd"

#define BUFFER_SIZE 4096

int logf(const char *format, ...);

int open_logfile()
{
	int log = open(LOGFILE, O_WRONLY | O_APPEND | O_CREAT, S_IRUSR | S_IWUSR);
	if(log == -1)
	{
		perror("open log file");
		exit(1);
	}
	dup2(log, 1);
	dup2(log, 2);
	close(log);
	logf("Log file opened\n");
	return 0;
}

void sigHUP(int signum)
{
	signal(SIGHUP, sigHUP);
	close(1);
	close(2);
	open_logfile();
}

void sigTERM(int signum)
{
	logf("Exiting\n");
	exit(0);
}

char *terminal = 0;

int logf(const char *format, ...)
{
	int len=0;
	time_t t;
	time(&t);
	char *ct = ctime(&t);
	ct[24] = 0;
	if(terminal)
	{
		len += fprintf(stderr, "%s [%d] %s: ", ct, getpid(), terminal);
	}
	else
	{
		len += fprintf(stderr, "%s [%d] ", ct, getpid());
	}

	va_list ap;
	va_start(ap, format);
	len += vfprintf(stderr, format, ap);
	va_end(ap);
	return len;
}

int strmatch(char *str, char *chars)
{
	// return 0 if str is made up only of chars in chars
	if(str)
	{
		if(chars)
		{
			int len = strlen(chars);
			int i=0;
			while(str[i] != 0)
			{
				// see if this char is OK
				int match = 0;
				for(int c = 0; c < len; c++)
				{
					if(str[i] == chars[c])
					{
						match++;
						break;
					}
				}
				if(match)
				{
					i++;
				}
				else
				{
					return 1;
				}
			}
			return 0; // seems to match
		}
		else
		{
			return 0;
		}
	}
	else
	{
		return 1;
	}
}

class Ras
{
private:
	int state;

	int quit;
	char *errorString;

	int fHandle;
	char device[256]; // this should be a constant...
	char *options_file;
	long speed;
	char *pppd;
	
	timeval timeout;
	int useTimeout;
	
	void setErrorMsg(char *msg)
	{
		//state = kError;
		errorString = msg;
	}

	void setError()
	{
		state = kError;
	}

	void setFatalError()
	{
		state = kFatalError;
	}

	int readSettings()
	{
	/*
		device = "/dev/ttyS1";
		speed = 115200;
		options_file = "/etc/ppp/options.ras1";
		*/
		if(speed == -1 || device == 0)
		{
			return rError;
		}
		else
		{
			return rOK;
		}
	}

	int openPort()
	{
		fHandle = open(device, O_RDWR | O_NOCTTY | O_NDELAY);
		if(fHandle == -1)
      {
			setErrorMsg("opening tty");
			return rError;
		}
		if(isatty(fHandle) == 0)
		{
			setErrorMsg("not a tty\n");
			return kError;
		}
		tcflush(fHandle, TCIFLUSH);

		fcntl(fHandle, F_SETFL, FNDELAY); // Configure port reading

		termios options;

		tcgetattr(fHandle, &options);
		cfsetispeed(&options, speed);
		cfsetospeed(&options, speed);


		options.c_cflag |= (CLOCAL | CREAD);
		options.c_cflag &= ~PARENB;
		options.c_cflag &= ~CSTOPB;
		options.c_cflag &= ~CSIZE;
		options.c_cflag |=  CS8;
		options.c_cflag |=  CRTSCTS;

		options.c_lflag &= ~(ICANON | ECHO | ISIG);

		tcsetattr(fHandle, TCSANOW, &options);

		return rOK;
	}

	void setTimeout(int seconds)
	{
		if(seconds >= 0)
		{
			useTimeout = 1;
			timeout.tv_sec = seconds;
			timeout.tv_usec = 0;
		}
		else
		{
			useTimeout = 0;
		}
	} 

	int readln(int handle, char *buffer, int len)
	{
		// read data up to a \n

		// use select to wait?
		int p = 0;
		while(p < len)
		{
			char c;


			int r = read(handle, &c, 1);
			if(r == 1)
			{
				if(c == '\n')
				{
					if(p > 0)
					{
						buffer[p++] = 0;
						break;
					}
				}
				else if(c == '\r')
				{
					// skip
				}
				else
				{
					buffer[p++] = c;
				}
			}
			else if(r == 0)
			{
				//logf(".");
			}
			else if(r == -1)
			{
				if(errno == EAGAIN)
				{
					fd_set rfds;
					timeval tv;
					FD_ZERO(&rfds);
					FD_SET(handle, &rfds);                      

					timeval *tvptr;
					if(useTimeout)
					{
						tv.tv_sec = timeout.tv_sec;
						tv.tv_usec = timeout.tv_usec;
						tvptr = &tv;
					}
					else
					{
						tvptr = 0;
					}

					int retval = select(handle + 1, &rfds, NULL, NULL, tvptr);
					if(retval == 1)
					{
						// we have data
					}
					else if(retval == 0)
					{
						// timeout
						return rTimeout;
					}
					else if(retval == -1)
					{
						// error
						return rError;
					}
				}
				else
				{
					perror("read");
					return rError;
				}
			}
		}
		return rOK;
	}

	int send(char *s)
	{
		return write(fHandle, s, strlen(s));
	}

	int chat(char *send, char *recv)
	{
		// send send and then wait up to timeout for recv.
		// if send is null, then don't send anything
		// if recv is null, then don't wait for anything

		if(send)
		{
			write(fHandle, send, strlen(send));
			write(fHandle, newLine, strlen(newLine));
		}

		if(recv)
		{

			char buffer[1024];
	
			while(true)
			{
				int r = readln(fHandle, buffer, sizeof(buffer));
				if(r == rOK)
				{
					logf("Read: (%d) %s\n", strlen(buffer), buffer);
					// does it match the requested string?
					if(strstr(buffer, recv) != 0)
					{
						break;
					}
				}
				else if(r == rTimeout)
				{
					return rTimeout;
				}
				else
				{
					return rError;
				};
			}
		}
		return rOK;
	}


public:

	Ras()
	{
		state = kStart;
		quit = 0;
		errorString = "none";
		device[0] = 0;
		options_file = 0;
		pppd = PPPD_DEFAULT;
		init = initString;
	}

/*	int setOptions(char *fDevice, char *fSpeed, char *fOptions)
	{
		device = fDevice;
		options_file = fOptions;
		speed = -1;
		if(strcmp(fSpeed, "1200")   ==0) speed = B1200;
		if(strcmp(fSpeed, "2400")   ==0) speed = B2400;
		if(strcmp(fSpeed, "4800")   ==0) speed = B4800;
		if(strcmp(fSpeed, "9600")   ==0) speed = B9600;
		if(strcmp(fSpeed, "19200")  ==0) speed = B19200;
		if(strcmp(fSpeed, "38400")  ==0) speed = B38400;
		if(strcmp(fSpeed, "115200") ==0) speed = B115200;
		return 0;
	}*/

	int setDevice(char *fDevice)
	{
		strncpy(device, fDevice, sizeof(device));
//		device = fDevice;
		terminal = device;
		return 0;
	}

	int setSpeed(char *fSpeed)
	{
		speed = -1;
		if(strcmp(fSpeed, "1200")   ==0) speed = B1200;
		if(strcmp(fSpeed, "2400")   ==0) speed = B2400;
		if(strcmp(fSpeed, "4800")   ==0) speed = B4800;
		if(strcmp(fSpeed, "9600")   ==0) speed = B9600;
		if(strcmp(fSpeed, "19200")  ==0) speed = B19200;
		if(strcmp(fSpeed, "38400")  ==0) speed = B38400;
		if(strcmp(fSpeed, "115200") ==0) speed = B115200;
		return 0;
	}

	char pppbuffer[1024];
	char *init;

	void setPPP(char *ppp)
	{
		logf("Setting pppd to %s\n", ppp);
		strncpy(pppbuffer, ppp, sizeof(pppbuffer));
		pppd = pppbuffer;
	}
	
	void setINIT(char *str)
	{
		logf("Setting init to %s\n", str);
		init = strdup(str);
	}
	
	int loadOptions(char *fOptions)
	{
		int device = 0;
		int speed = 0;
		int result = 0;

		options_file = fOptions;

		FILE *f = fopen(options_file, "r");
		if(f)
		{
			// read each line and look for a device and a speed...
			char buffer[BUFFER_SIZE];

			while(fgets(buffer, sizeof(buffer), f))
			{
				// remove any trailing \n
				int len = strlen(buffer) - 1;
				if(buffer[len] == '\n')
				{
					buffer[len] = 0;
				}
				// see if the line starts with "/dev/"
				if(strncmp(buffer, DEV_START, strlen(DEV_START)) == 0)
				{
					// this is the device
					setDevice(buffer);
					logf("Found device %s\n", buffer);
					device++;
				}
				if(strncmp(buffer, PPPD_START, strlen(PPPD_START)) == 0)
				{
					setPPP(buffer + strlen(PPPD_START));
				}
				if(strncmp(buffer, INIT_START, strlen(INIT_START)) == 0)
				{
					setINIT(buffer + strlen(INIT_START));
				}

				// and if it is a number with no ':'s then it is the speed
				if(len > 1 && strmatch(buffer, "1234567890\n") == 0)
				{
					setSpeed(buffer);
					logf("Found speed %s\n", buffer);
					speed++;
				}
			}
			if(speed == 0)
			{
				fprintf(stderr, "Failed to find a speed setting\n");
				result = 1;
			}
			else if(speed > 1)
			{
				fprintf(stderr, "Found more than one speed setting\n");
			}

			if(device == 0)
			{
				fprintf(stderr, "Failed to find a device setting\n");
				result = 1;
			}
			else if(device > 1)
			{
				fprintf(stderr, "Found more than one device setting\n");
			}
			fclose(f);
		}
		else
		{
			fprintf(stderr, "Configuration file %s\n", options_file);
			perror("Error opening configuration file");
				result = 1;
		}
		return result;
	}


	void doPPP()
	{
		int cpid = fork();
		if(cpid == 0)
		{
			// the child
			execl(pppd, pppd, "nodetach", device, "file" , options_file, 0);
			logf("error %d execing %s\n", errno, pppd);
			exit(21);
		}

		int status;
		waitpid(cpid, &status, 0);
		logf("pppd died with status %d\n", status);
	}

	void run()
	{
		while(!quit)
		{
			switch(state)
			{
				case kStart:
				{
					logf("Starting\n");

					// read settings
					if(readSettings())
					{
						state = kOpenPort;
					}
					else
					{
						setErrorMsg("Could not read settings\n");
						state = kFatalError;
					}
				}
				break;
	
				case kOpenPort:
				{
					logf("Opening port\n");

					if(openPort())
					{
						state = kInitModem;
					}
					else
					{
						perror("Open port");
						state = kError;
					}
				}
				break;
				case kInitModem:
				{
					logf("Init modem\n");
					setTimeout(5);
					if(chat(init, cmdOK) == rOK)
					{
						state = kWaitRing;
					}
					else
					{
						state = kHangup;
					}
				}
				break;
	
				case kWaitRing:
				{
					logf("Waiting for ring\n");
					setTimeout(-1);
					chat(0, cmdRING);
					state = kAnswer;
					// setError("?");
				}
				break;
	
				case kAnswer:
				{
					logf("Answering call\n");
					setTimeout(45);
					int r = chat(cmdANSWER, cmdCONNECT); 
					if(r == rOK)
					{
						state = kPPPd;
					}
					else if(r == rTimeout)
					{
						state = kHangup;
					}
					else
					{
						state = kError;
					}
				}
				break;
	
				case kPPPd:
				{
					logf("Starting ppp\n");
					doPPP();
					close(fHandle);

					state = kOpenPort;
					//state = kHangup;
				}
				break;
	
				case kHangup:
				{
					logf("Hanging up\n");
					//close(fHandle);
					//openPort();
					setTimeout(5);
					sleep(1);
					send("+++");
					sleep(1);
					if(chat(cmdHANGUP, cmdOK) == rOK)
					{
						logf("hung up\n");
						state = kInitModem;
					}
					else
					{
						logf("not responding\n");
						setErrorMsg("Modem is not responding");
						state = kError;
					}
				}
				break;
	
				case kError:
				{
					logf("Error : %s\n", errorString);
					close(fHandle);
					sleep(60);
					state = kStart;
					//quit = 1;
				}
				break;

				case kFatalError:
				{
					logf("Fatal error : %s\n", errorString);
					quit = 1;
				}
				break;
	
				default:
				{
					setErrorMsg("Unknown FSM state");
					state = kError;
				}
				break;
			}
		}
		logf("Quitting");
	}
};

int daemon_init()
{
	pid_t pid = fork();

	if(pid < 0)
	{
		return -1;
	}
	else if(pid != 0)
	{
		exit(0); // parent returns immediatly
	}
	close(0);
	close(1);
	close(2);

	// this is the child

	setsid();
	chdir("/");
	umask(0);
	return 0;
}

int main(int argc, char **argv)
{
	if(argc != 2)
	{
		fprintf(stderr, "Usage: %s optionsfile\n", argv[0]);
		return 1;
	}
	if(daemon_init() < 0)
	{
		perror("fork");
		exit(1);
	}
	else
	{
		open_logfile();
		// we need to set up signal catches...
		signal(SIGHUP, sigHUP);
		signal(SIGTERM, sigTERM);

		logf("RAS daemon version %s\n", VERSION);
		logf("Written by John Newbigin\n\n");
		if(argc == 2)
		{
			Ras *rasd = new Ras();
			rasd->loadOptions(argv[1]);
			
			rasd->run();
		}
		else
		{
		}
	}
}
